# ----------------------------------------------------------------------------
# SymForce - Copyright 2022, Skydio, Inc.
# This source code is under the Apache 2.0 license found in the LICENSE file.
# ----------------------------------------------------------------------------

from __future__ import annotations

import collections
import enum
import functools
import jinja2
import jinja2.ext
import os
from pathlib import Path
import textwrap

from symforce import logger
from symforce import typing as T
from symforce.codegen import format_util

CURRENT_DIR = os.path.dirname(__file__)
CPP_TEMPLATE_DIR = os.path.join(CURRENT_DIR, "cpp_templates")
PYTHON_TEMPLATE_DIR = os.path.join(CURRENT_DIR, "python_templates")
LCM_TEMPLATE_DIR = os.path.join(CURRENT_DIR, "lcm_templates")


class FileType(enum.Enum):
    CPP = enum.auto()
    PYTHON = enum.auto()
    PYTHON_INTERFACE = enum.auto()
    CUDA = enum.auto()
    LCM = enum.auto()
    MAKEFILE = enum.auto()
    TYPESCRIPT = enum.auto()

    @staticmethod
    def from_extension(extension: str) -> FileType:
        if extension in ("c", "cpp", "cxx", "cc", "tcc", "h", "hpp", "hxx", "hh"):
            return FileType.CPP
        elif extension in ("cu", "cuh"):
            return FileType.CUDA
        elif extension == "py":
            return FileType.PYTHON
        elif extension == "pyi":
            return FileType.PYTHON_INTERFACE
        elif extension == "lcm":
            return FileType.LCM
        elif extension == "Makefile":
            return FileType.MAKEFILE
        elif extension == "ts":
            return FileType.TYPESCRIPT
        else:
            raise ValueError(f"Could not get FileType from extension {extension}")

    @staticmethod
    def from_template_path(template_path: Path) -> FileType:
        parts = template_path.name.split(".")
        if parts[-1] != "jinja":
            raise ValueError(
                f"template must be of the form path/to/file.ext.jinja, got {template_path}"
            )
        return FileType.from_extension(parts[-2])


class RelEnvironment(jinja2.Environment):
    """
    Override join_path() to enable relative template paths. Modified from the below post.

    https://stackoverflow.com/questions/8512677/how-to-include-a-template-with-relative-path-in-jinja2
    """

    def join_path(self, template: T.Union[jinja2.Template, str], parent: str) -> str:
        return os.path.normpath(os.path.join(os.path.dirname(parent), str(template)))


def add_preamble(source: str, name: Path, filetype: FileType) -> str:
    prefix = (
        "//"
        if filetype in (FileType.CPP, FileType.CUDA, FileType.LCM, FileType.TYPESCRIPT)
        else "#"
    )
    dashes = "-" * 77
    return (
        textwrap.dedent(
            f"""
            {prefix} {dashes}
            {prefix} This file was autogenerated by symforce from template:
            {prefix}     {name}
            {prefix} Do NOT modify by hand.
            {prefix} {dashes}

            """
        ).lstrip()
        + source
    )


@functools.lru_cache
def jinja_env(template_dir: T.Openable) -> RelEnvironment:
    """
    Helper function to cache the Jinja environment, which enables caching of loaded templates
    """
    loader = jinja2.FileSystemLoader(os.fspath(template_dir))
    env = RelEnvironment(
        loader=loader,
        trim_blocks=True,
        lstrip_blocks=True,
        keep_trailing_newline=True,
        undefined=jinja2.StrictUndefined,
    )
    return env


def render_template(
    template_path: T.Openable,
    data: T.Dict[str, T.Any],
    output_path: T.Optional[T.Openable] = None,
    template_dir: T.Openable = CURRENT_DIR,
    autoformat: bool = True,
) -> str:
    """
    Boiler plate to render template. Returns the rendered string and optionally writes to file.

    Args:
        template_path: file path of the template to render
        data: dictionary of inputs for template
        output_path: If provided, writes to file
        template_dir: Base directory where templates are found, defaults to symforce/codegen
        autoformat: Run a code formatter on the generated code
    """
    logger.debug(f"Template  IN <-- {template_path}")
    if output_path:
        logger.debug(f"Template OUT --> {output_path}")

    if not isinstance(template_path, Path):
        template_path = Path(template_path)

    if not isinstance(template_dir, Path):
        template_dir = Path(template_dir)

    template_name = template_path.resolve().relative_to(template_dir.resolve())

    filetype = FileType.from_template_path(Path(template_name))

    template = jinja_env(template_dir).get_template(os.fspath(template_name))
    rendered_str = add_preamble(str(template.render(**data)), template_name, filetype)

    if autoformat:
        if filetype in (FileType.CPP, FileType.CUDA):
            # Come up with a fake filename to give to the formatter just for formatting purposes, even
            # if this isn't being written to disk
            if output_path is not None:
                format_cpp_filename = os.path.basename(output_path)
            else:
                format_cpp_filename = os.fspath(template_name).replace(".jinja", "")

            rendered_str = format_util.format_cpp(
                rendered_str, filename=os.path.join(CURRENT_DIR, format_cpp_filename)
            )
        elif filetype == FileType.PYTHON:
            rendered_str = format_util.format_py(rendered_str)
        elif filetype == FileType.PYTHON_INTERFACE:
            rendered_str = format_util.format_pyi(rendered_str)

    if output_path:
        directory = os.path.dirname(output_path)
        if not os.path.exists(directory):
            os.makedirs(directory)

        with open(output_path, "w") as f:
            f.write(rendered_str)

    return rendered_str


class TemplateList:
    """
    Helper class to keep a list of (template_path, output_path, data) and render
    all templates in one go.
    """

    TemplateListEntry = collections.namedtuple(
        "TemplateListEntry", ["template_path", "output_path", "data"]
    )

    def __init__(self) -> None:
        self.items: T.List = []

    def add(
        self, template_path: T.Openable, output_path: T.Openable, data: T.Dict[str, T.Any]
    ) -> None:
        self.items.append(
            self.TemplateListEntry(template_path=template_path, output_path=output_path, data=data)
        )

    def render(self, autoformat: bool = True) -> None:
        for entry in self.items:
            render_template(
                template_path=entry.template_path,
                output_path=entry.output_path,
                data=entry.data,
                autoformat=autoformat,
            )
